#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "funcs.h"
#include "config.h"
#include "db.h"

#define CHECK_URL(x) ( (sizeof(x) > 2 && x[sizeof(x) - 2] == '/') ? "" : "/" )

struct UNIQ_TAGS {
	char **tags;
	size_t size;
};

char *GetDate(const char *fmt, time_t time) {
	struct tm *tm = localtime(&time);

	static char buf[256];
	if (strftime(buf, sizeof(buf) - 1, fmt, tm) > 0)
		return buf;

	return "Unknown";
}

struct DB_STR *build_html_page(const struct CONFIG cfg, FILE *fp, const char *file, const struct UNIQ_TAGS ut, struct DB_STR *db, const size_t page, const size_t pages) {
	fprintf(fp, "<html>\n<head>\n<link rel='stylesheet' href='%s'>\n<title>%s</title>\n<meta charset='UTF-8'>\n</head>\n<body>\n<center><a href='index.html'><img id='logo' src='%s'></a>\n<h1 id='desc'>%s | <a href='%s'>RSS</a><br>%s/p%zu</h1></center>\n", CSS, TITLE, LOGO, DESC, XML_FILE, file, page);

	/* Tags */
	fputs("<center><div class='alltags'>\n<a href='index.html'>all</a>", fp);
	for (size_t i = 0; i < ut.size; i++)
		fprintf(fp, "<a href='%s.html'> %s</a>", ut.tags[i], ut.tags[i]);

	fputs("\n</center></div><br><br>\n", fp);

	/* Images */
	size_t images = 0;
	struct DB_STR *p = db;
	while (p != NULL) {
		struct DB_STR *i = p->next;
		if (strcmp(file, "index")) {
			char flag = 1;
			for (size_t i = 0; i < p->size; i++)
				if (!strcmp(p->tags[i], file))
					flag = 0;

			if (flag) {
				p = i;
				continue;
			}
		}

		fprintf(fp, "<a href='%s'>\n<img src='%s' alt='%s'></a><br>\n%s<br>\n", p->filename, p->filename, p->filename, GetDate(cfg.date_fmt, p->birthtime));
		for (size_t i = 0; i < p->size; i++)
			fprintf(fp, " • <a href='%s.html'>%s</a>", p->tags[i], p->tags[i]);

		if (p->desc)
			fprintf(fp, " %s", p->desc);

		fputs("\n<br><br><br>\n", fp);

		p = i;
		images++;
		if (images == cfg.posts_at_page)
			break;
	}

	fputs("<div class='pager'><center>\n", fp);
	for (size_t i = 0; i <= pages; i++) {
		if (i == 0) {
			if (i == page)
				fprintf(fp, "<a id='curr' href='%s.html'>0</a> ", file);

			else
				fprintf(fp, "<a href='%s.html'>0</a> ", file);
		}

		else {
			char *new_path = create_path("%s-%zu.html", file, i);

			if (i == page)
				fprintf(fp, "<a id='curr' href='%s'>%zu</a> ", new_path, i);

			else
				fprintf(fp, "<a href='%s'>%zu</a> ", new_path, i);
		}
	}

	fprintf(fp, "\n</center></div>\n</body>\n</html>\n");
	return p;
}

void build_html(const struct CONFIG cfg, const char *file, const struct UNIQ_TAGS ut, struct DB_STR *db) {
	size_t page = 0;
	size_t pages = 0;

	/* Count pages */
	size_t count = 0;
	struct DB_STR *p = db;
	while (p != NULL) {
		struct DB_STR *i = p->next;

		if (count > cfg.posts_at_page) {
			count = 0;
			pages++;
		}

		if (!strcmp(file, "index"))
			count++;

		else {
			for (size_t i = 0; i < p->size; i++)
				if (!strcmp(file, p->tags[i]))
					count++;
		}

		p = i;
	}

	p = db;
	while (1) {
		char *new_path;
		if (page == 0)
			new_path = create_path("%s.html", file);

		else
			new_path = create_path("%s-%zu.html", file, page);

		FILE *fp = file_open(cfg.prog_name, new_path, "w");
		if (fp == NULL)
			return;

		/* Builders */
		p = build_html_page(cfg, fp, file, ut, p, page, pages);
		if (p == NULL) {
			fclose(fp);
			break;
		}

		fclose(fp);
		page++;
	}
}


void build_xml(const struct CONFIG cfg, struct DB_STR *db) {
	FILE *fp = file_open(cfg.prog_name, cfg.xml_file, "w");
	if (fp == NULL)
		return;

	/* Header */
	fprintf(fp, "<?xml version='1.0' encoding='UTF-8' ?>\n<rss version='2.0' xmlns:atom='http://www.w3.org/2005/Atom'>\n<channel>\n\t<atom:link href='%s' rel='self' type='application/rss+xml' />\n\t<title>%s</title>\n\t<link>%s</link>\n\t<description>%s</description>\n\t<language>%s</language>\n", LINK, TITLE, LINK, DESC, LANGUAGE);

	/* Content */
	size_t count = 0;
	struct DB_STR *p = db;
	while (p) {
		struct DB_STR *i = p->next;

		/* Post */
		fprintf(fp, "\n<item>\n\t<title>%s</title>\n\t<link>%s%s%s</link>\n\t<guid>%s%s%s</guid>\n\t<pubDate>\n\t\t%s\t</pubDate>\n\t<description><![CDATA[\n\t\t<p><img src='%s%s%s' alt='plup'></p>\n\t\t<p>",
			p->filename,
			LINK, CHECK_URL(LINK), p->filename,
			LINK, CHECK_URL(LINK), p->filename,
			GetDate(cfg.date_fmt, p->birthtime),
			LINK, CHECK_URL(LINK), p->filename);

		/* Tags */
		for (size_t i = 0; i < p->size; i++)
		fprintf(fp, "<a href='%s%s%s.html'> %s</a>&nbsp; ",
			LINK, CHECK_URL(LINK), p->tags[i],
			p->tags[i]);

		fputs("</p>\n\t]]></description>\n</item>", fp);
		if (count == cfg.max_posts)
			break;

		p = i;
		count++;
	}

	fputs("\n</channel>\n</rss>\n", fp);
	fclose(fp);
}

char **uniq(struct DB_STR *db, size_t *uniq_size) {
	char **uniq_tag = malloc(sizeof(char *));
	if (uniq_tag == NULL)
		return NULL;

	struct DB_STR *p = db;
	while (p != NULL) {
		struct DB_STR *i = p->next;

		for (size_t i = 0; i < p->size; i++) {
			if (p->tags == NULL)
				break;

			char flag = 0;
			for (size_t j = 0; j < *uniq_size; j++) {
				if (!strcmp(p->tags[i], uniq_tag[j]))
					flag = 1;
			}

			if (flag)
				continue;

			char **bckp = realloc(uniq_tag, sizeof(char *) * (*uniq_size + 1));
			if (bckp == NULL) {
				free(uniq_tag);
				return NULL;
			}

			uniq_tag = bckp;
			uniq_tag[*uniq_size] = p->tags[i];
			(*uniq_size)++;
		}

		p = i;
	}

	return uniq_tag;
}

int rebuild(const struct CONFIG cfg) {
	struct DB_STR *db = read_db(cfg);
	if (db == NULL)
		return 1;

	/* Uniq */
	size_t uniq_size = 0;
	char **uniq_tag = uniq(db, &uniq_size);
	if (uniq_tag == NULL) {
		fprintf(stderr, "%s: malloc: %s\n", cfg.prog_name, strerror(errno));
		free_db(db);
		return 1;
	}

	struct UNIQ_TAGS ut = {.tags = uniq_tag, .size = uniq_size};
	for (size_t i = 0; i < uniq_size; i++)
		build_html(cfg, uniq_tag[i], ut, db);

	build_html(cfg, "index", ut, db);
	build_xml(cfg, db);

	/* Close */
	free(uniq_tag);
	free_db(db);
	return 0;
}
